import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.text.*;
import ij.measure.*;

/** This plugin performs a voxelwise Fuzzy C Mseans clustering on a set of datasets (images).
 *
 * @author Fred Damen <fred@damen.org>
 *
 * Version History
 * 2020-04-01 Released
 */
public class Fuzzy_C_Means implements PlugIn, DialogListener,
                                              MouseListener, 
                                              MouseMotionListener, 
                                              KeyListener {

   Vector<ImagePlus> imgVec = new Vector<ImagePlus>();
   Panel dspw = null;
   Font stepsFont = null;

   static File propFile = null;
   static Properties properties = (Properties)IJ.getProperty("FCM:Properties");


   /** The gui interface
    */
   public void run(String arg) {
      if (imgIds == null || imgIds.length < 1) {
         IJ.noImage();
         return;
         }

      gd = new GenericDialog("Fuzzy C Means");
      //gd.setFont(((new Dialog(IJ.getInstance())).getFont()));
      stepsFont = gd.getFont();
//IJ.log("font: "+stepsFont);
      stepsFont = stepsFont.deriveFont(Font.BOLD, stepsFont.getSize2D()*1.125F);

      gd.addMessage("Step 1: Choose the datasets(Images), viz. parameters", stepsFont);
      dspw = new Panel();
      dspw.setLayout(new GridBagLayout());
      gd.addPanel(dspw, GridBagConstraints.WEST, new Insets(0,15,0,0));
      addDatasetChoice(null);

      Vector<ResultsTable> rts  = new Vector<ResultsTable>();
      Choice rtc = null;
      Frame[] fs = WindowManager.getNonImageWindows();
      if (fs != null) {
         for(Frame f : fs)
            if (f.getClass().getName().endsWith("TextWindow") && ((TextWindow)f).getResultsTable() != null) {
               if (rtc == null) {
                  gd.addChoice("Preload Centers with ", new String[0],"");
                  rtc = (Choice)gd.getChoices().lastElement();
                  rtc.add("<none>");
                  rts.add(null);
                  }
               rtc.add(f.getTitle());
               rts.add(((TextWindow)f).getResultsTable());
               }
         if (rtc != null)
            rtc.select("Centers");
         }
     
      gd.addMessage("Necessary scaling will use Window/Level");

      if (properties == null) {
         properties = new Properties();
         IJ.setProperty("FCM:Properties",properties);
         }
      IJ.log("PS="+properties.size());
      gd.addButton(properties.size() == 0 ? "Load Properties " : "Properties Exist",new ActionListener() { 
         @Override public void actionPerformed(ActionEvent e) {

            OpenDialog od = new OpenDialog("Properties File");
            if (od.getFileName() != null)
               propFile = new File(od.getDirectory(),od.getFileName());
            try { properties.load(new FileInputStream(propFile)); } catch(Throwable t) {}
            IJ.log("ps="+properties.size());
            if (properties.size() > 0) {
               Button b = (Button)e.getSource();
               b.setLabel("Properties Exist"); 
               }
            } } );
      
      gd.showDialog();
      if (gd.wasCanceled()) return;

      ResultsTable rt = rtc != null ? rts.get(gd.getNextChoiceIndex()) : null;
      if (rt != null)
         for(String imgName : rt.getColumnAsStrings("Label")) {
            ImagePlus imp = WindowManager.getImage(imgName);
            if (imp == null) {
               IJ.error("Image "+imgName+" not currently open.");
               return;
               }
            imgVec.add(imp);
            }

      for(Component comp : dspw.getComponents()) {
         int c = ((Choice)comp).getSelectedIndex();
         if (c > 0)
            imgVec.add(WindowManager.getImage(imgIds[c-1]));
         }
 
      if (imgVec.isEmpty()) {
         IJ.error("No datasets selected.");
         return;
         }

      //loc = gd.getLocation();
      //IJ.setProperty("PreferedLocation",loc);

      processAndDisplay(rt);
      }

   int[] imgIds = WindowManager.getIDList();
   private void addDatasetChoice(Choice pc) {
      if (!( (dspw.getComponentCount() == 0) || (dspw.getComponent(dspw.getComponentCount()-1) == pc)) )
         return;

      if (imgIds == null || imgIds.length < 1)
         return;
      Choice c = new Choice();
      c.add("<choose>");
      for(int id : imgIds)
         c.add(WindowManager.getImage(id).getTitle());
      c.select(0);

      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridy = dspw.getComponentCount();
      dspw.add(c,gbc);
      dspw.revalidate();
      dspw.repaint();
      if (dspw.getComponentCount() > 1) gd.pack();

      c.addItemListener(new ItemListener() {
         @Override public void itemStateChanged(ItemEvent e) { addDatasetChoice((Choice)e.getSource()); } 
         });
      }

   static void saveProperties() {
      String dn = propFile == null ? "" : propFile.getParent();
      String fn = propFile == null ? "" : propFile.getName();
      SaveDialog sd = new SaveDialog("Properties file",dn,fn,"properties");
      if (sd.getDirectory() == null || sd.getFileName() == null)
         return;
      propFile = new File(sd.getDirectory(),sd.getFileName());
      if (saveProperties(propFile,"FCM:maps"))
         IJ.log("Properties saved");
      }
   static boolean saveProperties(File pf, String ... names) {
      Properties pp = new Properties();
      for(String n : names) 
         if (IJ.getProperty(n) instanceof Properties) {
            IJ.log("n="+n);
            Properties p = (Properties)IJ.getProperty(n);
            String np = n.split(":",2)[1];
            for(String pn : p.stringPropertyNames())
               pp.setProperty(n+":"+pn,p.getProperty(pn));
         }
      IJ.log("pf="+pf);
      try { pp.store(new FileWriter(pf),"Fuzzy_C_Means Properties"); }
      catch(Throwable t) { 
         StringWriter sw = new StringWriter();
         t.printStackTrace(new PrintWriter(sw));
         IJ.log("Error FileWriter "+sw.toString());
         return false; }
      return true;
      }

   static float mean(ImageProcessor ip, Roi roi) {
      int n = 0;
      float sum = 0F;
      float[][] arr = ip.getFloatArray();
      for(Point p : roi) {
         float v = arr[p.x][p.y];
         if (Float.isFinite(v)) {
            sum += v;
            n++;
            }
         }
      return sum/n;
      }

   TextField[] cursorValues = null;
   TextField[] matchedValues = null;
   TextField[] withinValues = null;
   private TextField[] createValues(String name, int col) {
      int ni = imgVec.size();
      TextField[] values = new TextField[ni+1+1];

      Panel p = new Panel();
      GridBagConstraints pc = new GridBagConstraints();
      pc.fill = GridBagConstraints.NONE;
      pc.anchor = GridBagConstraints.NORTH;
      pc.gridx = col;
      pc.gridy = 0;
      pc.gridheight = ni+1+1;
      leftPanel.add(p,pc);
      p.setLayout(new GridBagLayout());

      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.HORIZONTAL;
      c.gridx = 0;

      values[0] = new TextField(name,8);
      //values[0].setFont(new Font("SanSerif", Font.PLAIN, 19));
      values[0].setEditable(false);
      c.gridy = 0;
      p.add(values[0],c);
      
      for(int i=0; i<ni; i++) {
         values[i+1] = new TextField("NaN");
         //values[i+1].setFont(new Font("SanSerif", Font.PLAIN, 19));
         values[i+1].setEditable(false);
         c.gridy = i+1;
         p.add(values[i+1],c);
         }

      values[values.length-1] = new TextField("likely",8);
      //values[values.length-1].setFont(new Font("SanSerif", Font.PLAIN, 19));
      values[values.length-1].setEditable(false);
      c.gridy = values.length-1;
      p.add(values[values.length-1],c);
      return values;
      }
   private void updateCursorValues(int ix, int iy, int iz, String current) {
      if (imgVec == null) return;
      int ni = imgVec.size();
      for(int i=0; i<ni; i++)
         try { cursorValues[i+1].setText(String.format("%.3g", imgVec.get(i).getStack().getProcessor(iz).getPixelValue(ix,iy))); }
         catch(Exception e) { cursorValues[i+1].setText("UGH"); }
      if (resVec == null) return;
      int nm = cenVec.size();
      int mi = -1;    Integer[] mia = new Integer[nm];
      float mv = -1f; float[]   mva = new float[nm];
      int ci = -1;
      Font f = cenVec.get(0)[ni+1+1-1].getFont();
      for(int i=0; i<nm; i++) {
//IJ.log("resVec("+i+")="+resVec.get(i));
         float v = Float.NaN; try { v = resVec.get(i).getStack().getProcessor(iz).getPixelValue(ix,iy); } catch(Exception e) {}
         if (v > mv) {
            mi = i;
            mv = v;
            }
         cenVec.get(i)[ni+1+1-1].setText(String.format("%.3g",v));
         cenVec.get(i)[ni+1+1-1].setFont(f.deriveFont(v<0.05 ? Font.PLAIN : Font.BOLD));
         cenVec.get(i)[ni+1+1-1].setBackground(v<0.02 ? null : Color.white);
         cenVec.get(i)[ni+1+1-1].setForeground(resVec.get(i).getTitle().equals(current) ? Color.red : Color.black);
         if (resVec.get(i).getTitle().equals(current))
            ci = i;

         mia[i] = i;
         mva[i] = v;
         }
      cursorValues[ni+1+1-1].setText((mi==-1)?"None":cenVec.get(mi)[0].getText());
      for(int i=0; i<matchedValues.length; i++)
         matchedValues[i].setText(cenVec.get(mi)[i].getText());
      matchedValues[ni+1+1-1].setFont(f.deriveFont(matchedValues[0].getText().equals(current) ? Font.BOLD : Font.PLAIN));
      cenVec.get(mi)[ni+1+1-1].setForeground(Color.blue);

      for(int i=0; i<withinValues.length; i++)
         withinValues[i].setText((ci==-1) ? "" : cenVec.get(ci)[i].getText());
      withinValues[0].setForeground((ci==-1) ? Color.black : cenVec.get(ci)[0].getForeground());

      for(int i=0; i<colors.length; i++) try {
         colorFields[i].setText("NaN");
         float v = WindowManager.getImage(Colors.colorToString(colors[i])).getStack().getProcessor(iz).getPixelValue(ix,iy); 
         colorFields[i].setText(String.format("%.3g",v));
         } catch(Exception e) {};
      try {
         String str = "("+ix+","+iy+","+iz+(ci>-1?String.format(":%2.0f",resVec.get(ci).getDisplayRangeMax()*100):"")+") ";
         for(Integer i : sortedIndexes(mia, mva))
            if (mva[i] >= 0.01)
               str += cenVec.get(i)[0].getText()+String.format("/%2.0f, ",mva[i]*100);
         cenOrder.setText(str);
         } catch(Exception e) { cenOrder.setText("oops:"+e); }
      }

   private static Integer[] sortedIndexes(Integer[] ia, float[] va) {
      Arrays.sort(ia, new Comparator<Integer>() {
         @Override public int compare(final Integer o1, final Integer o2) {
            return Float.compare(va[o2], va[o1]);
            }
         });
      return ia;
      }

   Vector<TextField[]> cenVec = new Vector<TextField[]>();
   private void addCenter(Roi roi) {
      if (daccb!=null && daccb.getState()) IJ.log("daccb=true");
      if (daccb!=null && daccb.getState()) return;
      if (roi == null || roi.getState() != Roi.NORMAL)
         return;

IJ.log("addCenter="+roi);

      int iz = roi.getImage().getZ();
//      Rectangle bounds = roi.getBounds();
//      int ix = bounds.x;
//      int iy = bounds.y;
      
      int ni = imgVec.size();
      float[] arr = new float[ni];
      for(int i=0; i<ni; i++)
         arr[i] = mean(imgVec.get(i).getStack().getProcessor(iz),roi);
         //arr[i] = imgVec.get(i).getStack().getProcessor(iz).getPixelValue(ix,iy);
//IJ.log("before = "+cenVec.get(0)[0].getFont());
      addCenter(null, arr);
//IJ.log("after = "+cenVec.get(0)[0].getFont());
      gd.pack();
      }
   Font cenFont = null;
   private void addCenter(String name, float[] arr) {
      if (daccb!=null && daccb.getState()) IJ.log("daccb=true");
      if (daccb!=null && daccb.getState()) return;

      int nc = cenVec.size();

      TextField[] center = new TextField[imgVec.size()+1+1];

      Panel p = new Panel();
      GridBagConstraints pc = new GridBagConstraints();
      pc.gridx = nc; //+1+3;
      pc.gridy = 0;
      pc.gridheight = imgVec.size()+1+1;
      centersPanel.add(p,pc);
      p.setLayout(new GridBagLayout());

      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.HORIZONTAL;
      c.gridx = 0;

      center[0] = new TextField(name!=null ? name : "C"+(nc+1),8); 
      center[0].setFont(cenFont);     
      center[0].setForeground(colorMap != null && colorMap.containsKey(center[0].getText()) ? colorMap.get(center[0].getText()) : null);
      center[0].addKeyListener(new KeyListener() {
         @Override public void keyPressed(KeyEvent e) {}
         @Override public void keyTyped(KeyEvent e) {}
         @Override public void keyReleased(KeyEvent e) {
            if (!e.isControlDown()) return;
            TextField tf = (TextField)e.getSource();
            if (e.getKeyCode() == KeyEvent.VK_R) {
               String oldName = tf.getText();
               String newName = IJ.getString("New Name:",oldName);
               if (newName.isEmpty()) return;
               if (findResImp(newName) != null) {
                  IJ.error("Name '"+newName+"' already exists.");
                  return;
                  }
               tf.setText(newName);
               ImagePlus rimp = findResImp(oldName);
               if (rimp == null) return;
               rimp.setTitle(newName);
               Point p = locMap.get(oldName);      locMap.remove(oldName);   locMap.put(newName, p);
               Dimension d = sizeMap.get(oldName); sizeMap.remove(oldName);  sizeMap.put(newName,d);
               Color c = colorMap.get(oldName);    colorMap.remove(oldName); colorMap.put(newName,c);
               }
            if (e.getKeyCode() == KeyEvent.VK_F) {
//IJ.log("tf = "+tf.getText());
               ImagePlus imp = findResImp(tf.getText());
               IJ.selectWindow(imp.getID());
               ImageWindow win = imp.getWindow();
               if (win != null && e.isShiftDown()) {
                  Point p = tf.getLocationOnScreen();
                  p.translate(tf.getWidth(),0);
                  win.setLocation(p);
                  if (e.isAltDown())
                     IJ.run(imp, "Enhance Contrast", "saturated=0.35");
                  }
               }
            if (e.getKeyCode() == KeyEvent.VK_D)
               if (e.isShiftDown() && e.isAltDown()) {
                  Panel p  = (Panel)tf.getParent();
                  Panel pp = (Panel)p.getParent();
                  pp.remove(p);
                  for(TextField[] tfa : cenVec)
                     if (tfa[0].equals(tf)) {
                        cenVec.remove(tfa);
                        resVec.remove(findResImp(tf.getText()));
                        IJ.showStatus("Removed "+tf.getText());
                        IJ.log("Removed "+tf.getText());
                        break;
                        }
                  //IJ.log("delete ME ");
                  gd.pack();
                  }
            if (e.getKeyCode() == KeyEvent.VK_S) 
               if (!e.isShiftDown() && !e.isAltDown()) {
                  for(TextField[] tfa : cenVec)
                     if (tfa[0].equals(tf)) 
                        for(TextField t : tfa)
                           if (t != tf)
                              t.setEnabled(!t.isEnabled());
                  }
               else if (e.isShiftDown() && !e.isAltDown()) {
                  for(TextField[] tfa : cenVec)
                     if (tfa[0].equals(tf)) {
                        for(TextField t : tfa)
                           if (t != tfa[0])
                              t.setEnabled(true);
                        }
                     else {
                        for(TextField t : tfa)
                           if (t != tfa[0])
                              t.setEnabled(false);
                        }
                  
                  }
               else if (e.isAltDown()) {
                  for(TextField[] tfa : cenVec)
                     for(TextField t : tfa)
                        if (t != tfa[0])
                           t.setEnabled(true);
                  }
            if (e.getKeyCode() == KeyEvent.VK_M) {
               Color tfc = tf.getForeground();
               if (!e.isShiftDown() && !e.isAltDown()) {
                       if (!tf.isForegroundSet())   tf.setForeground(Color.red);
                  else if (tfc.equals(Color.red))   tf.setForeground(Color.green);
                  else if (tfc.equals(Color.green)) tf.setForeground(Color.blue);
                  else if (tfc.equals(Color.blue))  tf.setForeground(null);
                  else                              tf.setForeground(Color.red);
                  }
               else if (!e.isShiftDown() && e.isAltDown()) {
                  ImagePlus resimp = findResImp(tf.getText()).duplicate();
                  IJ.run(resimp, "Set...", "value=0 stack");
                  for(TextField[] tfa : cenVec)
                     if (tfa[0].getForeground().equals(tfc))
                        (new ImageCalculator()).run("Add 32-bit stack", resimp, findResImp(tfa[0].getText()));
                  resimp.setDefaultDisplayRange(0.0,1.0);
                  resimp.setDisplayRange(0.0,1.0);
                  resimp.setTitle(Colors.colorToString2(tfc));
                  resimp.show();
                  resimp.changes = false;
                  engage(resimp);
                  }
               }
            }
         });

      center[0].addMouseMotionListener(new MouseMotionListener() {
            @Override public void mouseDragged(MouseEvent e) {}
            @Override public void mouseMoved(MouseEvent e) {
               TextField   tf = (TextField)e.getSource();
               if (e.isShiftDown()) {
                  Panel       p  = (Panel)tf.getParent();
                  Component[] cs = p.getComponents();
                  for(int i=0; i<matchedValues.length; i++)
                     matchedValues[i].setText(((TextField)cs[i]).getText());
                  }
               if (e.isControlDown()) {
                  ImagePlus imp = findResImp(tf.getText());
                  if (imp == null) {
                     IJ.showStatus(tf.getText()+" not found","flash");
                     return;
                     }
                  IJ.selectWindow(imp.getID());
                  ImageWindow win = imp.getWindow();
                  if (win != null && e.isShiftDown()) {
                     Point p = tf.getLocationOnScreen();
                     p.translate(tf.getWidth(),0);
                     win.setLocation(p);
                     if (e.isAltDown())
                        IJ.run(imp, "Enhance Contrast", "saturated=0.35");
                     }
                  }
               }
            });

      c.gridy = 0;
      p.add(center[0],c);
      
      for(int i=0; i<arr.length; i++) {
         center[i+1] = new TextField(String.format("%.3g",arr[i]),8);
         center[i+1].setFont(cenFont); 
         c.gridy = i+1;
         p.add(center[i+1],c);
         }
      for(int i=arr.length; i<imgVec.size(); i++) {
         center[i+1] = new TextField("NaN",8);
         c.gridy = i+1;
         p.add(center[i+1],c);
         }
      for(int i=0; i<center.length-1; i++)
         center[i].addFocusListener(new FocusListener() {
                     @Override public void focusGained(FocusEvent e) {((TextField)e.getSource()).selectAll();}
                     @Override public void focusLost(FocusEvent e) {((TextField)e.getSource()).select(0,0);}
                     });

      center[center.length-1] = new TextField("",8);
      center[center.length-1].setFont(cenFont); 
      center[center.length-1].setEditable(false);
      c.gridy = center.length-1;
      p.add(center[center.length-1],c);
      try { centersPanelSize.setSize(sizeOf(nc)); } catch(Throwable t) {IJ.log("oops "+t);}

      center[0].requestFocus();
      cenVec.add(center);
      gd.pack();
      }

   ResultsTable ssrt = null;
   private void reloadCenters(ResultsTable rt) {
      if (rt == null) {
         IJ.error("No Centers snapshot");
         return;
         }
      
      for(TextField[] tfa : cenVec) {
         Panel p  = (Panel)tfa[0].getParent();
         Panel pp = (Panel)p.getParent();
         pp.remove(p);
         }
      cenVec.clear();

      for(int i=0; i<=rt.getLastColumn(); i++)
         if (!rt.getColumnHeading(i).equals("Label"))
            addCenter(rt.getColumnHeading(i),rt.getColumn(i));

      for(TextField[] tfa : cenVec)
         tfa[0].setForeground(ColorGroupMap.get(tfa[0].getText()));

      gd.pack();
      }

   Dimension centersPanelSize = new Dimension(0,0);
   ScrollPane csp = new ScrollPane(ScrollPane.SCROLLBARS_ALWAYS) { @Override public Dimension getPreferredSize() { return centersPanelSize;} };
   private Dimension sizeOf(int n) {
      Panel p = new Panel();
      p.setLayout(new GridBagLayout());
      
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.HORIZONTAL;
      for(int i=0; i<imgVec.size()+1+1; i++) {
      TextField tf = new TextField("",8);
      tf.setFont(cenFont);
         c.gridy = i;
         p.add(tf,c);
         }
      Frame f = new Frame();
      f.setLayout(new BorderLayout());
      f.add(p, BorderLayout.CENTER);
      f.pack();
      Dimension pis = p.getSize();
      f.remove(p);
      f.dispose();

      pis.width  += 4;
      pis.height += 20;

      Dimension size = new Dimension(pis.width*n,pis.height);

      Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
      int maxWidth = 7 * screen.width / 8;
      int maxHeight = 3 * screen.height / 4;
      if (size.width > maxWidth) size.width = (maxWidth/pis.width)*pis.width;
      if (size.height > maxHeight) size.height = maxHeight;

      //size.width += 50;
      //size.height += 25;
//IJ.log("size = "+size);

      return size;
      }



   TreeMap<String,Color> ColorGroupMap = new TreeMap<String,Color>();
   Color[] colors = new Color[]{Color.red, Color.green, Color.blue};
   Panel mainPanel = new Panel();
   Panel leftPanel = new Panel();
   Panel centersPanel = new Panel();
   Panel colorPanel = new Panel();
   TextField[] colorFields;
   Label cenOrder = new Label("",Label.LEFT);
   GenericDialog gd;
   TextField[] imgNames;
   TextField cfmtw = null;
   Checkbox daccb = null;
   Scrollbar zs = null;
   TextField zsn = null;
   Checkbox processWidget = null;
   //Checkbox snapshotWidget = null;
   //Checkbox snapshotReloadWidget = null;
   private void processAndDisplay(ResultsTable rt) {
      int ni = imgVec.size();

      gd = new NonBlockingGenericDialog("Fuzzy C Means");
      gd.addDialogListener(this);

      gd.addMessage("Step 2: Configure the centers ...",stepsFont);
      gd.addMessage("Click on representitve voxels with Roi\nMay subsitute min/mid/max for values.");
//panel.setFont(gd.getFont().deriveFont(16f));
mainPanel.setFont(new Font("SanSerif", Font.PLAIN, 17));
IJ.log("panel font: "+mainPanel.getFont());
cenFont = mainPanel.getFont();
      gd.addPanel(mainPanel,GridBagConstraints.WEST, new Insets(0, 0, 0, 0));
      mainPanel.setLayout(new GridBagLayout());
      {
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.NORTHWEST;
      c.gridx = 0;
      c.gridy = 0;
      c.weightx = 1.0;
      mainPanel.add(leftPanel,c);
      leftPanel.setLayout(new GridBagLayout());
      }
      {
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.NORTH;
      c.gridx = 0;
      c.gridy = 0;
      c.weightx = 0;
      TextField tf = new TextField("Image / Cluster",20);
      tf.setEditable(false);
      leftPanel.add(tf,c);
      }

      scale = new double[ni][2];
      imgNames = new TextField[ni];
      for(int i=0; i<ni; i++) {
         scale[i] = new double[]{imgVec.get(i).getDisplayRangeMin(), imgVec.get(i).getDisplayRangeMax()-imgVec.get(i).getDisplayRangeMin()};

         GridBagConstraints c = new GridBagConstraints();
         c.fill = GridBagConstraints.NONE;
         c.anchor = GridBagConstraints.NORTH;
         c.gridx = 0;
         c.gridy = i+1;
         c.weightx = 0;
         imgNames[i] = new TextField(imgVec.get(i).getTitle(),20);
         imgNames[i].setEditable(true);

         imgNames[i].addKeyListener(new KeyListener() {
            @Override public void keyPressed(KeyEvent e) {}
            @Override public void keyTyped(KeyEvent e) {}
            @Override public void keyReleased(KeyEvent e) {
               if (!e.isControlDown()) return;
               TextField tf = (TextField)e.getSource();
               if (e.getKeyCode() == KeyEvent.VK_S) {
                  if (!e.isShiftDown() && !e.isAltDown())
                     tf.setBackground(tf.getBackground()==Color.white ? Color.lightGray : Color.white);
                  else if (e.isShiftDown() && !e.isAltDown())
                     for(TextField t : imgNames)
                        t.setBackground(t.equals(tf) ? Color.white : Color.lightGray);
                  else if (e.isAltDown())
                     for(TextField t : imgNames)
                        t.setBackground(Color.white);
                  }
               }
            });
         imgNames[i].setBackground(Color.white);

         leftPanel.add(imgNames[i],c);

         imgVec.get(i).getWindow().getCanvas().addMouseListener(this);
         imgVec.get(i).getWindow().getCanvas().addMouseMotionListener(this);
         imgVec.get(i).getWindow().getCanvas().addKeyListener(this);
//IJ.log(imgVec.get(i).getTitle());
         }

      {
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.NORTH;
      c.gridx = 0;
      c.gridy = ni+1;
      c.weightx = 0;
      TextField tf = new TextField("Cursor Memberships",20);
      tf.setEditable(false);
      leftPanel.add(tf,c);
      }

      cursorValues  = createValues("Cursor",  1);
      matchedValues = createValues("Matched", 2);
      withinValues  = createValues("Within",  3);


      for(String n : new String[]{"GM","GM1","GM2","GM3","GM4","GM5","WM","WM1","WM2","WM3","WM4","WM5","LB1","LB2",
                                  "WMH1","WMH2","MS3","C26","C28","C31","CB1",
                                  "BV1","BV2"})
         ColorGroupMap.put(n,Color.green);
      for(String n : new String[]{"Lesion1","Lesion2","Lesion3","Lesion4","Lesion5","Lesion6","Lesion7",
                                  "Lesion10","Lesion11","Lesion12","Lesion13","Lesion14","Lesion15","Lesion16",
                                  "Lesion8","Lesion9","Lesion10","C27","WMH1","WMH2","WMH3","C30","C31","C32","WMH4","C34","C35","C36","C37","C38","C39","C40",
                                  "TUMOR1","TUMOR2","TUMOR3","TUMOR4","TUMOR5","TUMOR6","TUMOR7","TUMOR8","TUMOR9","TUMOR10","TUMOR11",
                                  "TUMOR1A","TUMOR1B","TUMOR1C","TUMOR1D","TUMOR2A","TUMOR2B","TUMOR2C","TUMOR2D",
                                  "EP1","EDEMA1","EDEMA2","EDEMA3","EDEMA4"})
         ColorGroupMap.put(n,Color.red);
      for(String t : new String[]{"TUMOR","EDEMA"})
         for(int i=0; i<10; i++)
            if (i==0) ColorGroupMap.put(t+i,Color.red);
            else for(String a : new String[]{"A","B","C","D","E","F","G","H","I"})
                    ColorGroupMap.put(t+i+a,Color.red);
      for(String n : new String[]{"Fluid1","Fluid2","Fluid3","Flow1","Flow2","Flow3",
                                  "FLUID1","FLUID2","FLUID3","FLOW1","FLOW2","FLOW3","FLOW4"})
         ColorGroupMap.put(n,Color.blue);


      {
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.NORTHWEST;
      c.gridx = 0;
      c.gridy = 0;
      c.weightx = 0.0;
      csp.add(centersPanel,c);
      centersPanel.setLayout(new GridBagLayout());
      }
      if (true) {
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.NORTHWEST;
      c.gridx = 1;
      c.gridy = 0;
      c.gridheight = ni+1+1;
      mainPanel.add(csp,c);
      }

      ssrt = rt;
/*
      if (rt != null)
         for(int i=0; i<=rt.getLastColumn(); i++)
            if (!rt.getColumnHeading(i).equals("Label"))
               addCenter(rt.getColumnHeading(i),rt.getColumn(i));

      for(TextField[] tfa : cenVec)
         tfa[0].setForeground(ColorGroupMap.get(tfa[0].getText()));
*/
      reloadCenters(ssrt);

      gd.addPanel(colorPanel,GridBagConstraints.WEST, new Insets(0, 0, 0, 0));
      colorPanel.setLayout(new GridBagLayout());
      colorFields = new TextField[3];
      for(int i=0; i<colors.length; i++ ) {
         colorFields[i] = new TextField(Colors.colorToString(colors[i]),20);
         colorFields[i].setForeground(colors[i]);
         colorFields[i].setBackground(Color.white);
         GridBagConstraints c = new GridBagConstraints();
         c.fill = GridBagConstraints.HORIZONTAL;
         c.gridx = i;
         c.gridy = 0;
         c.weighty = 0.5;
         colorFields[i].setEditable(false);
         colorPanel.add(colorFields[i],c);
         }
      {
         //cenOrder.setMaximumSize(new Dimension(Integer.MAX_VALUE,Integer.MAX_VALUE));
         cenOrder.setText(String.format("%350s","x"));
         GridBagConstraints c = new GridBagConstraints();
         c.fill = GridBagConstraints.HORIZONTAL;
         c.gridx = 3;
         c.gridy = 0;
         c.weighty = 0.5;
         colorPanel.add(cenOrder,c);
         }
      
      int cbi = 0;

      gd.addCheckbox("Done adding centers",false);
      daccb = (Checkbox)gd.getCheckboxes().lastElement(); cbi++;

      gd.addSlider("Z",1,imgVec.get(0).getNSlices(),imgVec.get(0).getZ());
      zs = (Scrollbar)gd.getSliders().lastElement();
      zs.addAdjustmentListener(new AdjustmentListener() {
         public void adjustmentValueChanged(AdjustmentEvent e) {
            int z = ((Scrollbar)e.getSource()).getValue();
            scrollAll(z);
            }
         });
      zs.addMouseWheelListener(new MouseWheelListener() {
         public void mouseWheelMoved(MouseWheelEvent e) {
            int z = ((Scrollbar)e.getSource()).getValue();
            scrollAll(z);
            }
         });
      zs.setPreferredSize(new Dimension(200,30));
      zsn = (TextField)gd.getNumericFields().lastElement();

      gd.addMessage("Settings...",stepsFont);
      gd.addNumericField("Fuzzy Factor:", 2);
      gd.addNumericField("Threshold:",1e-4,10);
      gd.addNumericField("Iterations:", 1,0);
      gd.addNumericField("Found:",Double.NaN);
      cfmtw = (TextField)gd.getNumericFields().lastElement();
      gd.addCheckbox("Close All before processing",true); 
      cbi++;

      gd.addMessage("Step 3: Run the Fuzzy Clustering ...",stepsFont);
      gd.addCheckbox("Process                          ", false);
      processWidget = (Checkbox)gd.getCheckboxes().lastElement(); cbi++;
/*
      gd.addCheckbox("Centers snapshot",false);
      snapshotWidget = (Checkbox)gd.getCheckboxes().lastElement(); cbi++;
      gd.addCheckbox("Centers snapshot reload",false);
      snapshotReloadWidget = (Checkbox)gd.getCheckboxes().lastElement(); cbi++;
      if (ssrt == null) snapshotReloadWidget.setEnabled(false);
*/

      gd.addCheckboxGroup(1,3,new String[]{"Centers snapshot","Sorted", "Reload"},new boolean[]{false, false, false});
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { showCenters(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { showSortedCenters(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { reloadCenters(ssrt); ((Checkbox)e.getSource()).setState(false);} });

      gd.addCheckboxGroup(1,2,new String[]{"Save setup","Save Properties"},new boolean[]{false,false});
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { saveSetup(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { saveProperties(); ((Checkbox)e.getSource()).setState(false);} });

      gd.addCheckboxGroup(1,4,new String[]{"Conglomerate","Row", "Grid", "Sumation"},new boolean[]{false, false, false, false});
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { conglomerate1(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { conglomerate2(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { conglomerate3(); ((Checkbox)e.getSource()).setState(false);} });
      ((Checkbox)gd.getCheckboxes().get(cbi++)).addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) { conglomerate4(); ((Checkbox)e.getSource()).setState(false);} });

      gd.setCancelLabel("Just Leave");
      gd.setOKLabel("Leave & Close'em");
      if (locMap.containsKey("Fuzzy_C_Means"))
         gd.setLocation(locMap.get("Fuzzy_C_Means"));
      gd.showDialog();
      IJ.showStatus("Fuzzy C Means: Closing");
      disengage();
      for(int i=0; i<ni; i++) try {
          imgVec.get(i).getWindow().getCanvas().removeMouseListener(this);
          imgVec.get(i).getWindow().getCanvas().removeMouseMotionListener(this);
          imgVec.get(i).getWindow().getCanvas().removeKeyListener(this);
          } catch(Exception e) {}
      imgVec.clear();
      imgVec = null;
      if (gd.wasCanceled()) return;

      IJ.showStatus("Fuzzy C Means: Closing All");
      IJ.wait(100);
      closeAll();
      IJ.showStatus("Fuzzy C Means: Closed All");
      }

   private void scrollAll(int z) {
      if (resVec != null)
         for(ImagePlus imp : resVec) {
            imp.setZ(z);
            if (imp.getType() ==  ImagePlus.GRAY32 && !imp.isComposite() && imp.getProp("noWL") == null) {
               IJ.resetMinAndMax(imp);
               IJ.run(imp, "Enhance Contrast", "saturated=0.35");
               }
            }
      }

   MouseWheelListener mwlhandler = new MouseWheelListener() {
      @Override public void mouseWheelMoved(MouseWheelEvent e) {
         ImagePlus imp = ((ImageWindow)e.getSource()).getImagePlus();
         IJ.setKeyUp(KeyEvent.VK_SHIFT);
         imp.getWindow().mouseWheelMoved(e);
         if (!e.isShiftDown())
            return;
   
         int z = imp.getZ();
         zs.setValue(z);
         zsn.setText(""+z);
         scrollAll(z);
         }
      };

   Vector<ImagePlus> resVec = new Vector<ImagePlus>();
   MouseMotionListener resMML = new MouseMotionListener() {
            @Override public void mouseDragged(MouseEvent e) {}
            @Override public void mouseMoved(MouseEvent e) {
               if ((e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) == MouseEvent.SHIFT_DOWN_MASK)
                  return;
               ImageCanvas ic = (ImageCanvas)e.getSource();
               Point p = ic.getCursorLoc();
               updateCursorValues(p.x,p.y,ic.getImage().getZ(),ic.getImage().getTitle());
               }
            };
   private void engage(ImagePlus... imps) {
      for(ImagePlus imp : imps) {
         resVec.add(imp);
         imp.getWindow().getCanvas().addMouseMotionListener(resMML);
         imp.getWindow().removeMouseWheelListener(imp.getWindow());
         imp.getWindow().addMouseWheelListener(mwlhandler);
         }
      }
   private void disengage() {
      for(ImagePlus imp : resVec)
         try { 
            imp.getWindow().getCanvas().removeMouseMotionListener(resMML); 
            imp.getWindow().removeMouseWheelListener(mwlhandler);
            imp.getWindow().addMouseWheelListener(imp.getWindow());
            } catch(Exception e) {}
      }
   private void closeAll() {
      for(ImagePlus imp : resVec)
         if (imp != null && imp.getWindow() != null)
            imp.getWindow().close();
      resVec.removeAllElements();
      }
   private ImagePlus findResImp(String str) {
      for(ImagePlus imp : resVec)
         if (imp.getTitle().equals(str))
            return imp;
      return null;
      }

   TreeMap<String,Point> locMap = IJ.getProperty("FCM:Locations")!=null ? (TreeMap<String,Point>)IJ.getProperty("FCM:Locations") : new TreeMap<String,Point>();
   TreeMap<String,Dimension> sizeMap = IJ.getProperty("FCM:Sizes")!=null ? (TreeMap<String,Dimension>)IJ.getProperty("FCM:Sizes") : new TreeMap<String,Dimension>();
   TreeMap<String,Color> colorMap = IJ.getProperty("FCM:Colors")!=null ? (TreeMap<String,Color>)IJ.getProperty("FCM:Colors") : new TreeMap<String,Color>();
   private void saveSetup() {
      locMap.clear();
      sizeMap.clear();
      for(ImagePlus imp : resVec) {
         if (imp != null || imp.getWindow() == null)
            continue;
         locMap.put(imp.getTitle(),imp.getWindow().getLocation());
         sizeMap.put(imp.getTitle(),imp.getWindow().getSize());
         }
      for(String name : new String[]{"Lesions", "Normal", "One of these Things..."}) 
         try { locMap.put(name, WindowManager.getImage(name).getWindow().getLocation()); } catch(Exception e) {}
      locMap.put("Fuzzy_C_Means",gd.getLocation());
      IJ.setProperty("FCM:Locations",locMap);
      IJ.setProperty("FCM:Sizes",sizeMap);

      colorMap.clear();
      for(TextField[] tfa : cenVec)
         colorMap.put(tfa[0].getText(),tfa[0].getForeground());
      }

   private void showMembershipMaps(double[][] m) {
      int nc = cenVec.size();
      ImagePlus[] cenimps = new ImagePlus[nc];
      ImagePlus tmpimp = imgVec.get(0).duplicate();
      IJ.run(tmpimp, "Set...", "value=NaN stack");
      for(int c=0; c<nc; c++) {
         cenimps[c] = tmpimp.duplicate();
         cenimps[c].setTitle(cenVec.get(c)[0].getText());
         }

      int nx = imgVec.get(0).getWidth();
      int ny = imgVec.get(0).getHeight();
      int nz = imgVec.get(0).getNSlices();
      int nf = imgVec.get(0).getNFrames();

      int d = 0;
      for(int[] xyz : xyzVec) {
         int c = 0;
         for(ImagePlus cenimp : cenimps)
            cenimp.getStack().getProcessor(xyz[2]+1).putPixelValue(xyz[0],xyz[1],m[c++][d]);
         d++;
         }
      Point loc = gd.getLocation();
      Dimension gdws = gd.getSize();
      loc.y += gdws.height;
      Dimension ws = imgVec.get(0).getWindow().getSize();
      double mag = imgVec.get(0).getWindow().getCanvas().getMagnification();
      int i = 0;
      ImagePlus maximp = cenimps[0].duplicate();
      for(ImagePlus imp : cenimps) {
         ImagePlus resimp = findResImp(imp.getTitle());
         if (resimp != null && resimp.getWindow() == null) resimp = null;

         Point tmploc = new Point(loc.x+i++*ws.width,loc.y);
         if (resimp != null) tmploc = resimp.getWindow().getLocation();
         if (locMap.containsKey(imp.getTitle()))
            tmploc = locMap.get(imp.getTitle());
/*
         if (resimp != null) ImageWindow.setNextLocation(resimp.getWindow().getLocation());
                        else ImageWindow.setNextLocation(loc.x+i++*ws.width,loc.y);
         if (locMap.containsKey(imp.getTitle()))
            ImageWindow.setNextLocation(locMap.get(imp.getTitle()));
*/

         int z = (resimp!=null) ? resimp.getZ() : imgVec.get(0).getZ();
         Dimension tmpws = resimp!=null ? resimp.getWindow().getSize() : ws;
         if (sizeMap.containsKey(imp.getTitle()))
            tmpws = sizeMap.get(imp.getTitle());

         show(imp, imp.getTitle(), 0.0, 1.0, z, tmpws, tmploc);
/*
         imp.setDefaultDisplayRange(0.0,1.0);
         imp.setDisplayRange(0.0,1.0);
         imp.show();

      imp.waitTillActivated();
      ImageWindow win = imp.getWindow();
      for(int ii=0; ii<100; ii++)
         if (win.isActive()) break;
         else { 
 //           IJ.log(imp.getTitle()+" not active yet"); 
            IJ.wait(3);
            }
      ImageCanvas ic = win.getCanvas();
      for(int ii=0; ii<100; ii++)
         if (ic.isValid()) break;
         else { IJ.log("C"+i); IJ.wait(10); }
      for(int ii=0; ii<100; ii++)
         if (ic.getWidth() != imp.getWidth() || ic.getHeight() != imp.getHeight()) {
            IJ.wait(1); ic.zoom100Percent(); IJ.wait(1);
            }

         imp.setZ((resimp!=null) ? resimp.getZ() : imgVec.get(0).getZ());

         imp.getWindow().setSize(resimp!=null ? resimp.getWindow().getSize() : ws);
         if (sizeMap.containsKey(imp.getTitle()))
            imp.getWindow().setSize(sizeMap.get(imp.getTitle()));

         imp.getWindow().getCanvas().fitToWindow();
*/

         (new ImageCalculator()).run("Max 32-bit stack", maximp, imp);
         resVec.remove(resimp);
         if (resimp != null) resimp.close();
         }
      //IJ.run(maximp, "Macro...", "code=if(v>"+cfmtw.getText()+")v=NaN;");
      IJ.run(maximp, "Macro...", "code=v=1.0-v; stack");
      maximp.setDisplayRange(0,1);
      maximp.setTitle("One of these Things...");
      tmpimp = findResImp(maximp.getTitle());
      if (tmpimp != null && tmpimp.getWindow() == null) tmpimp = null;
      if (tmpimp != null) ImageWindow.setNextLocation(tmpimp.getWindow().getLocation());
                     else ImageWindow.setNextLocation(loc.x+i++*ws.width,loc.y);
      maximp.show();
      maximp.getWindow().setSize(tmpimp!=null ? tmpimp.getWindow().getSize() : ws);
      maximp.getWindow().getCanvas().fitToWindow();
      //if (tmpimp != null) maximp.setZ(tmpimp.getZ());
      int z = (tmpimp!=null) ? tmpimp.getZ() : imgVec.get(0).getZ();
      maximp.setZ(z);
      maximp.changes = false;
      resVec.remove(tmpimp);
      if (tmpimp != null) tmpimp.close();

      Vector<ImagePlus> cimpv = new Vector<ImagePlus>();
      for(Color c : new Color[]{Color.red, Color.green, Color.blue}) try {
         if (locMap.containsKey(Colors.colorToString(c)))
            ImageWindow.setNextLocation(locMap.get(Colors.colorToString(c)));
         ImagePlus resimp = findResImp(Colors.colorToString(c));
         if (resimp != null && resimp.getWindow() == null) resimp = null;
         if (resimp != null) ImageWindow.setNextLocation(resimp.getWindow().getLocation());

         ImagePlus imp = Images_Calculator.sum(getImages(c));
         imp.setTitle(Colors.colorToString(c));
         imp.setDisplayRange(0.0,1.0);
         imp.setProp("noWL","true");
         imp.show();
         imp.waitTillActivated();
         Dimension cws = ws;
         if (resimp != null) 
            cws = resimp.getWindow().getSize();
         if (sizeMap.containsKey(Colors.colorToString(c))) 
            cws =sizeMap.get(Colors.colorToString(c));
         imp.getWindow().setSize(cws);
         imp.getWindow().getCanvas().fitToWindow();
         
         imp.setZ(z);
         cimpv.add(imp);
         //engage(imp);
         resVec.remove(resimp);
         if (resimp != null) resimp.close();

         if (false && c == Color.red) {
            File dir = (File)IJ.getProperty("DMS");
            IJ.saveAsTiff(imp,dir.getPath()+File.separator+"red.tiff");
            imp.setTitle("red");
            }
         } catch(Throwable t) {}

      closeAll();
      engage(cenimps);
      engage(maximp);

      engage(cimpv.toArray(new ImagePlus[0])); cimpv.clear();

      for(ImagePlus imp : cenimps)
         IJ.run(imp, "Enhance Contrast", "saturated=0.35");
      }

   private void show(ImagePlus imp, String title, double drmin, double drmax, int z, Dimension _ws, Point _loc) {
      ImageWindow.setNextLocation(_loc);
      imp.setDefaultDisplayRange(drmin,drmax);
      imp.setDisplayRange(drmin,drmax);
      imp.show();
      imp.waitTillActivated();
      ImageWindow win = imp.getWindow();
      for(int ii=0; ii<100; ii++)
         if (win.isActive()) break;
         else { 
 //           IJ.log(imp.getTitle()+" not active yet"); 
            IJ.wait(3);
            }
      ImageCanvas ic = win.getCanvas();
      for(int ii=0; ii<100; ii++)
         if (ic.isValid()) break;
         else { IJ.log("C"+ii); IJ.wait(10); }
      for(int ii=0; ii<100; ii++)
         if (ic.getWidth() != imp.getWidth() || ic.getHeight() != imp.getHeight()) {
            IJ.wait(1); ic.zoom100Percent(); IJ.wait(1);
            }

      imp.setZ(z);

      imp.getWindow().setSize(_ws);
      imp.getWindow().getCanvas().fitToWindow();
      }

   private String[] getImages(Color c) {
      Vector<String> titles = new Vector<String>();
      for(TextField[] tfa : cenVec)
         if (tfa[0].isForegroundSet() && c.equals(tfa[0].getForeground()))
            titles.add(tfa[0].getText());
      return titles.toArray(new String[0]);
      }

   private void conglomerate1() {

      GenericDialog gd = new GenericDialog("Conglomerate Maps");

      Vector<String> mcv  = new Vector<String>();
      mcv.add("<choose>");
      if (IJ.getProperty("FCM:maps") == null)
         IJ.setProperty("FCM:maps",new Properties());
      Properties mprop = (Properties)IJ.getProperty("FCM:maps");
      for(String key : mprop.stringPropertyNames())
         mcv.add(key);
      
      gd.addChoice("Choices:", mcv.toArray(new String[0]), "");
      Choice nc = (Choice)gd.getChoices().lastElement();
      gd.addStringField("Name: ","",20);
      TextField nt = (TextField)gd.getStringFields().lastElement();

      gd.addStringField("Title:","",20);
      TextField tt = (TextField)gd.getStringFields().lastElement();

      Panel p = new Panel();
      p.setLayout(new GridBagLayout());
      gd.addPanel(p,GridBagConstraints.NORTH, new Insets(0, 0, 0, 0));

      int[] ids = WindowManager.getIDList();
      Vector<String> itv = new Vector<String>();
      itv.add("<blank>");

      for(String s : WindowManager.getImageTitles())
         itv.add(s);

      Choice[] matrix = new Choice[7];
      for(int i=0; i<matrix.length; i++) {
         matrix[i] = new Choice();
         for(String title : itv)
            matrix[i].add(title);
         //matrix[i].select(maps[i]);
         GridBagConstraints c = new GridBagConstraints();
         c.fill = GridBagConstraints.VERTICAL;
         c.gridx = i;
         c.gridy = 0;
         c.weightx = 0.5;
         p.add(matrix[i],c);
         }

      nc.addItemListener(new ItemListener() { 
         @Override public void itemStateChanged(ItemEvent e) {
            Choice c = (Choice)e.getSource();
            if (c.getSelectedIndex() < 1) return;
            String n = c.getSelectedItem();
            nt.setText(n);

            String mapstr = mprop.getProperty(n);
            if (mapstr == null) return;
            String[] maps = mapstr.split(";");
            if (maps.length > 0) tt.setText(maps[0]);
            for(int i=1; i<maps.length; i++) {
               if (WindowManager.getImage(maps[i]) == null && IJ.showMessageWithCancel​(maps[i]+" does not exist","Create "+maps[i]+"?")) {
                  ImagePlus imp = Images_Calculator.sum(Images_Calculator.re2names(maps[i]));
                  if (imp == null) return;
                  imp.setTitle(maps[i]);
                  imp.setDefaultDisplayRange(0.0,imp.getDisplayRangeMax());
                  imp.setDisplayRange(0.0,imp.getDisplayRangeMax());
                  imp.show();
                  imp.setZ(zs.getValue());
                  IJ.run(imp, "Enhance Contrast", "saturated=0.35");
                  matrix[i-1].add(maps[i]);
                  }
               matrix[i-1].select(maps[i]);
               }
            } 
         });

      gd.addCheckbox("Row ?",true);
      gd.addNumericField("Font Fudge Factor", fontFudgeFactor);

      RoiManager roiManager = RoiManager.getInstance();
      Roi[] rois = null;;
      if (roiManager != null) {
         rois = roiManager.getRoisAsArray();
         Vector<String> rtv = new Vector<String>();
         rtv.add("<none>");
         for(Roi roi : rois)
            rtv.add(roi.getName());
         gd.addChoice("Background Roi", rtv.toArray(new String[0]),"");
         gd.addChoice("Foreground Roi", rtv.toArray(new String[0]),"");
         }

      gd.showDialog();
      if (gd.wasCanceled()) return;

      int ti = gd.getNextChoiceIndex();
      String template = ti == 0 ? null : mcv.get(ti);
      String name = gd.getNextString();
      String title = gd.getNextString();

      String str = title;
      Vector<String> mv = new Vector<String>();
      for(int i=0; i<matrix.length; i++)
         if (matrix[i].getSelectedIndex() > 0) {
            String v = matrix[i].getSelectedItem();
            mv.add(v);
            str += ";"+v;
            }
      if (mv.isEmpty()) return;
      mprop.setProperty(name,str);

      String[] maps = mv.toArray(new String[0]);
      IJ.log(name + " = " + str);

      ImagePlus imp = conglomerate(maps);
      imp.setTitle(title);
      imp.show();
      imp.setZ(zs.getValue());
      if (locMap.containsKey(name))  imp.getWindow().setLocation(locMap.get(name));
      if (sizeMap.containsKey(name)) imp.getWindow().setSize(sizeMap.get(name));
      imp.getWindow().getCanvas().fitToWindow();
      engage(imp);

      if (gd.getNextBoolean()) {

         Roi bgroi = null;
         Roi fgroi = null;

         if (rois != null) {
            int bri = gd.getNextChoiceIndex();
            if (bri > 0)
               bgroi = rois[bri-1];
            int fri = gd.getNextChoiceIndex();
            if (fri > 0)
               fgroi = rois[fri-1];
            }

         fontFudgeFactor = gd.getNextNumber();
         ImagePlus gimp = conglomerate(title, maps, bgroi, fgroi);
         if (gimp != null)
            gimp.show();
         }

      return;
      }

   private static ImagePlus conglomerate(String[] images) {
      for(String img : images) {
         if (img == null)
            continue;
         ImagePlus ugh = WindowManager.getImage(img); 
         if (ugh == null) {
            IJ.log("conglomerate: "+img+" missing.");
            return null;
            }
         ugh = ugh.duplicate();
         ugh.setTitle(img+"UGH");
         try { ugh.setDisplayRange(0.0,1.0); } catch(Throwable t) {IJ.log("IJ.resetMinAndMax is unhappy "+img);}
         ugh.show();
         }

      String str = "";
      for(int i=0; i<images.length; i++)
         if (images[i] != null)
            str += "c"+(i+1)+"="+images[i]+"UGH ";
      try { IJ.run("Merge Channels...", str+" create keep"); } catch(Throwable t) {IJ.log("Merge Channels... is unhappy "+str+"\n"+t); return null; }
      ImagePlus imp = WindowManager.getImage("Composite");

      for(String img : images)
         if (img != null)
            WindowManager.getImage(img+"UGH").close();

      for(int m=0,c=0; m<images.length; m++)
         if (images[m] != null) {
            for(int f=0; f<imp.getNFrames(); f++)
               for(int s=0; s<imp.getNSlices(); s++)
                  imp.getStack().setSliceLabel(images[m],imp.getStackIndex(c+1,s+1,f+1));
            c++;
            }

      return imp;
      }


   private void conglomerate2() {
      IJ.log("left="+leftPanel.getSize()+" centers="+centersPanel.getSize()+" gd="+gd.getSize()+" sizeOf(1)="+sizeOf(1));
      int soo = sizeOf(1).width - 4;
      int ncw = Math.min(cenVec.size(),(gd.getSize().width - leftPanel.getSize().width) / soo);
      IJ.log("ncw="+ncw+" "+(ncw*soo));
      centersPanelSize.width = ncw*soo + 8;
      csp.setSize(centersPanelSize);
      gd.pack();
      }

   Panel ccpw = null;
   private void conglomerate3() {
      GenericDialog gd = new GenericDialog("Conglomerate Conglomerate Maps");

      gd.addMessage("Rows...");
      ccpw = new Panel();
      ccpw.setLayout(new GridBagLayout());
      gd.addPanel(ccpw, GridBagConstraints.WEST, new Insets(0,15,0,0));
      addConglomerateChoice(null);

      gd.addNumericField("Font Fudge Factor", fontFudgeFactor);

      gd.showDialog();
      if (gd.wasCanceled()) return;

      fontFudgeFactor = gd.getNextNumber();

      Vector<ImagePlus> vvimps = new Vector<ImagePlus>();
      Properties mprop = (Properties)IJ.getProperty("FCM:maps");
      for(Component c : ccpw.getComponents()) {
         String cn = ((Choice)c).getSelectedItem();
         if (cn.equals("<choose>"))
            continue;
         String[] maps = mprop.getProperty(cn).split(";");
         String title = maps[0];
         maps = Arrays.copyOfRange(maps,1,maps.length);
         Roi bgroi = null, fgroi = null;
         for(String m : maps)
            if (WindowManager.getImage(m) == null) {
               ImagePlus imp = Images_Calculator.sum(Images_Calculator.re2names(m));
               if (imp == null) {
                  IJ.error("Could not create needed summation "+m);
                  return;
                  }
               imp.setTitle(m);
               double max = IJ.getNumber("Max for "+m, imp.getDisplayRangeMax());
               imp.setDefaultDisplayRange(0.0,max);
               imp.setDisplayRange(0.0,max);
               imp.show();
               imp.setZ(zs.getValue());
               }

         vvimps.add(conglomerate(title, maps, bgroi, fgroi));
         }

      if (vvimps.isEmpty()) return;
      ImagePlus cimp = The_Combiner.combineVertically(vvimps.toArray(new ImagePlus[0]));
      if (cimp != null) {
         cimp.setTitle("Conglomerate Conglomerate");
         cimp.show();
         }

      ccpw = null;
      }

   private void addConglomerateChoice(Choice pc) {
      if (ccpw == null)
         return;

      if (!( (ccpw.getComponentCount() == 0) || (ccpw.getComponent(ccpw.getComponentCount()-1) == pc)) )
         return;

      if (IJ.getProperty("FCM:maps") == null)
         return;

      Choice c = new Choice();
      c.add("<choose>");

      Properties mprop = (Properties)IJ.getProperty("FCM:maps");
      for(String key : mprop.stringPropertyNames())
         c.add(key);

      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridy = ccpw.getComponentCount();
      ccpw.add(c,gbc);
      ccpw.revalidate();
      ccpw.repaint();
      if (ccpw.getComponentCount() > 1) ((GenericDialog)ccpw.getParent()).pack();

      c.addItemListener(new ItemListener() {
         @Override public void itemStateChanged(ItemEvent e) { addConglomerateChoice((Choice)e.getSource()); } 
         });
      }



   private void conglomerate4() {
      String re = IJ.getString("Regex:","");
      if (re.isEmpty()) return;
      if (WindowManager.getImage(re) != null) {
         if (IJ.getString("Close "+re+" ?","yes").equals("yes"))
            WindowManager.getImage(re).close();
         else
            return;
         }
      if (re.isEmpty()) return;
      ImagePlus imp = Images_Calculator.sum(Images_Calculator.re2names(re));
      if (imp == null) return;
      imp.setTitle(re);
      imp.setDefaultDisplayRange(0.0,imp.getDisplayRangeMax());
      imp.setDisplayRange(0.0,imp.getDisplayRangeMax());
      imp.show();
      imp.setZ(zs.getValue());
      }

   private double fontFudgeFactor = 1.0;
   private ImagePlus conglomerate(String name, String[] images, Roi bgroi, Roi fgroi) {
      ImagePlus[] imps = new ImagePlus[images.length+1];
      int z = zs.getValue(); // WindowManager.getImage(name).getZ();
      //imps[0] = WindowManager.getImage(name).duplicate();
      ImagePlus ugh = conglomerate(images);
      ugh.setTitle(name+"-1");
      //imps[0].show();
      imps[0] = ugh.duplicate();
      ugh.close(); ugh= imps[0];
      imps[0].setZ(z);
      int w = imps[0].getWidth();
      int h = imps[0].getHeight();
      IJ.run(imps[0], "RGB Color", "slices");
      String roiColor = "red"; //Colors.colorToString(fgroi.getStrokeColor() != null ? fgroi.getStrokeColor() : Color.green);
      double fff = (bgroi != null) ? bgroi.getBounds().getHeight()/h : 1.0;
      Font font = new Font("Arial",Font.BOLD,(int)(4*16*fff*fontFudgeFactor));
      imps[0].show();
      imps[0] = new ImagePlus("",imps[0].getProcessor());
      ugh.close();
      if (fgroi != null) {
         ImageProcessor ip = imps[0].getProcessor();
         ip.setColor(roiColor);
         imps[0].setRoi(fgroi);
         fgroi.setStrokeColor(Color.green);
         fgroi.drawPixels(imps[0].getProcessor());
         IJ.run(imps[0], "Select None", "");
         }
      double wm = 4;
      double hm = 4;
      if (bgroi != null) {
         imps[0].setRoi(bgroi);
         imps[0] = imps[0].crop();
         wm *= imps[0].getWidth()/(double)w;
         hm *= imps[0].getHeight()/(double)h;
         }
      imps[0] = imps[0].resize(imps[0].getWidth()*4,imps[0].getHeight()*4,1,"none");
      ImageProcessor ip = imps[0].getProcessor();
      ip.setColor(Color.white);
      ip.setBackgroundColor(Color.black);
      ip.setFont(font);
      Rectangle nr = ip.getStringBounds(name);
      //ip.setFontSize(16);
//IJ.log("wm="+wm+" hm="+hm+" h="+nr.height);
      ip.drawString(name,(int)(10*wm), (int)(1.5*nr.height)); // nr.height+40); //(int)(20*hm+nr.height));
//IJ.log("UGH");
      for(int i=0; i<images.length; i++) {
         if (images[i] == null || images[i].isEmpty())
            continue;
         ImageProcessor fip = WindowManager.getImage(images[i]).getProcessor();
         imps[i+1] = Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",fip),bgroi, new ImagePlus("",fip),fgroi,
            " "+images[i],Double.NaN,Double.NaN,100,"<"+roiColor+",1>","400%x400%;00,00","None","Bottom-Left","Right","10%","%.0f%%",font,CompositeImage.colors[i],Color.black);
         IJ.run(imps[i+1],"RGB Color", "");
         ImageProcessor cip = imps[i+1].getProcessor();
         cip.setColor(CompositeImage.colors[i]);
         cip.setBackgroundColor(Color.black);
         cip.setFont(font);
         //cip.setFontSize(16);
         String s = String.format("max %3.0f%%",fip.getStats().max*100);
         Rectangle sr = cip.getStringBounds(s);
         cip.drawString(s,cip.getWidth()-sr.width-0,(int)(1.5*sr.height));
         }
      ImagePlus imp = The_Combiner.combineHorizontally(imps);
      imp.setTitle(name+" Colorized");
      //imp.show();
      return imp;
      }

   Vector<int[]> xyzVec = null;
   double[][] scale = null;
   private double[][] getData(double found) {
      processWidget.setLabel("Getting Data");
      Vector<double[]> dVec = new Vector<double[]>();
      int ni = imgVec.size();
      int nx = imgVec.get(0).getWidth();
      int ny = imgVec.get(0).getHeight();
      int nz = imgVec.get(0).getNSlices();
      int nf = imgVec.get(0).getNFrames();

      xyzVec = new Vector<int[]>();

      for(int z=0; z<nz; z++)
         for(int x=0; x<nx; x++)
      loop: for(int y=0; y<ny; y++) {
               if (Double.isFinite(found))
                  for(int c=0; c<cenVec.size(); c++) 
                     if (cenVec.get(c)[1].isEnabled()) {
                        double mv = resVec.get(c).getStack().getProcessor(z+1).getPixelValue(x,y);
                        if (Double.isNaN(mv) || mv <= found)
                           continue loop;
                        }
               double[] da = new double[ni];
               for(int i=0; i<ni; i++)
                  da[i] = scale(imgVec.get(i).getStack().getProcessor(z+1).getPixelValue(x,y), scale[i]);
               //if (!Double.isNaN(arrSum(da))) {
               if (!Double.isNaN(da[0])) {
                  dVec.add(da);
                  xyzVec.add(new int[]{x,y,z});
                  }
               }
      return dVec.toArray(new double[0][ni]);
      }

   private double[][] getCenters() {
      processWidget.setLabel("Getting Centers");
      int nc = cenVec.size();
      int ni = imgVec.size();
      if (nc<1 || ni<1) return null;
      double[][] centers = new double[nc+1][ni+1];
      for(int c=0; c<nc; c++)
         for(int i=0; i<ni; i++)
            switch(cenVec.get(c)[i+1].getText().toLowerCase()) {
               case "min": centers[c][i] = 0.0; break;
               case "mid": centers[c][i] = 0.5; break;
               case "med": centers[c][i] = 0.5; break;
               case "max": centers[c][i] = 1.0; break;
               default:    try {centers[c][i] = scale(Double.parseDouble(cenVec.get(c)[i+1].getText()), scale[i]); }
                           catch(Throwable e) {IJ.error("Center "+c+" '"+cenVec.get(c)[0].getText()+"' entry "+i+" is bogus '"+cenVec.get(c)[i+1].getText()+"'");}
               }
      for(int c=0; c<nc; c++)
         centers[c][ni] = (!cenVec.get(c)[1].isEnabled()) ? Double.NaN : 0.0;
      for(int i=0; i<ni; i++)
         centers[nc][i] = (imgNames[i].getBackground()!=Color.white) ? Double.NaN : 0.0;
      processWidget.setLabel("Computing...");
      return centers;
      }

   private void putCenters(double[][] centers) {
      for(int c=0; c<cenVec.size(); c++)
         for(int i=0; i<cenVec.get(c).length-1-1; i++)
            cenVec.get(c)[i+1].setText(String.format("%.3g",unscale(centers[c][i],scale[i])));
      }


   private static Integer[] sortedIndexes(Vector<TextField[]> va) {
      Integer[] ia = new Integer[va.size()];
      Arrays.setAll(ia, i -> i);
      Arrays.sort(ia, new Comparator<Integer>() {
         @Override public int compare(final Integer o1, final Integer o2) {
            return va.get(o1)[0].getText().compareTo(va.get(o2)[0].getText());
            }
         });
      return ia;
      }

   private void showSortedCenters() {
      showCenters(sortedIndexes(cenVec));
      }
   private void showCenters() {
      Integer[] ia = new Integer[cenVec.size()];
      Arrays.setAll(ia, i -> i);
      showCenters(ia);
      }
   private void showCenters(Integer[] ia) {
      if (cenVec.size() < 1) return;
      ssrt = new ResultsTable(cenVec.get(0).length-1-1);
      ssrt.setPrecision(6);
      for(int c=0; c<cenVec.size(); c++) {
         for(int i=0; i<cenVec.get(ia[c]).length-1-1; i++) {
            if (c == 0)
               ssrt.setLabel(imgNames[i].getText(),i);
            ssrt.setValue(cenVec.get(ia[c])[0].getText(),i, Double.parseDouble(cenVec.get(ia[c])[i+1].getText()));
            }
         }
      ssrt.show("Centers");
      //snapshotReloadWidget.setEnabled(true);
      }

   public static void logArr(String str, double[][] arr) {
      IJ.log(str);
      for(int i=0; i<arr.length; i++) {
         String s = "";
         for(int j=0; j<arr[i].length; j++)
            s += ", " + String.format("%5.2g", arr[i][j]);
         IJ.log(""+i+"  "+s.substring(1));
         }
      }

   public static void logArr(String str, double[] arr) {
      IJ.log(str);
      for(int i=0; i<arr.length; i++)
         IJ.log(""+i+"  "+arr[i]);
      }

   public static double scale(double v, double[] s) {
      return Math.min(1.0, Math.max(0.0, (v - s[0])/s[1]));
      }
   public static double unscale(double v, double[] s) {
      return v * s[1] + s[0];
      }

   public static double arrSum(double[] da) {
      double sum=0;
      for(int i=0; i<da.length; i++)
         sum += da[i];
      return sum;
      }

   public static String[] strArrCat(String s, String[] sa) {
      String[] res = new String[1+sa.length];
      res[0] = s;
      for(int i=0; i<sa.length; i++)
         res[i+1] = sa[i];
      return res;
      }

   /** Internal - do not call.
    */
   public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
      //if (e == null) IJ.log("e is null");
      int z = (int)gd.getNextNumber();
      double ff  = gd.getNextNumber();
      double eps = gd.getNextNumber();
      int    NI  = (int)gd.getNextNumber();
      double found = gd.getNextNumber();
      boolean closem = gd.getNextBoolean();

      if (processWidget.getState()) {
         processWidget.setLabel("Processing");
         double[][][] ugh = fcm(getData(found), getCenters(), ff, NI, eps);
         if (ugh == null) {
            IJ.error("fcm returned null");
            return false;
            }
         processWidget.setLabel("Done computing");
         //closeAll();
         showMembershipMaps(ugh[0]);
         putCenters(ugh[1]);
         processWidget.setState(false);
         if (NI > 1) {
            processWidget.setLabel("Plotting "+ugh[2][0].length+" Iterations");
            Point loc = gd.getLocation();
            Dimension ws = gd.getSize();
            ImageWindow.setNextLocation(loc.x+ws.width, loc.y+0*ws.height);
            Plot plot = new Plot("Goodness Of Fit","Iterations","Goodness of Fit");
            plot.setWindowSize(ws.width, ws.height-50);
            plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
            plot.setLineWidth(3);
            plot.setMaxIntervals(ugh[2][0].length);
            plot.addPoints(ugh[2][1],ugh[2][0],Plot.LINE);
            plot.show();
            engage(plot.getImagePlus());
            }
         processWidget.setLabel("Procesed "+ugh[2][0].length+" Iterations");
         }
/*
      if (snapshotWidget.getState()) {
         showCenters();
         }
      snapshotWidget.setState(false);

      if (snapshotReloadWidget.getState()) {
         reloadCenters(ssrt);
         }
      snapshotReloadWidget.setState(false);
*/
      //gd.pack();
      return true;
      }

//////
    // these listeners are activated if the selection is changed in the corresponding ImagePlus
    public synchronized void mousePressed(MouseEvent e) {  }   
    public synchronized void mouseDragged(MouseEvent e) {  }
    public synchronized void mouseClicked(MouseEvent e) { addCenter(((ImageCanvas)(e.getSource())).getImage().getRoi()); }
    public synchronized void keyPressed(KeyEvent e)     { /* addCenter(((ImageCanvas)(e.getSource())).getImage().getRoi()); */ }
    // unused listeners concering actions in the corresponding ImagePlus
    public void mouseReleased(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseMoved(MouseEvent e) {}
    public void keyTyped(KeyEvent e) {}
    public void keyReleased(KeyEvent e) {}

//////

   /** Perform the Fuzzy C Means.
    *
    * All data is assumed to be prenormalized.
    * @param data     objects x object's properties
    * @param centers  centers x center's locations
    * @param ff       fuzzy factor
    * @param NI       max number of iterations
    * @param eps      stop when goodness of fit delta is below eps
    * @return [0][][] objects x object's memberships
    *         [1][][] centers x object's center locations
    *         [2][][] good of fit per iteration, iteration number
    */
   public static double[][][] fcm(double[][] data, double[][] centers, double ff, int NI, double eps) {
      //processWidget.setLabel("Computing");
      if (data == null    || data.length    == 0) { IJ.error("data: "+data);       return null; }
      if (centers == null || centers.length == 0) { IJ.error("centers: "+centers); return null; }
      double[][] C = centers;
      boolean[][] fixed = null;
      if (centers[0].length > data[0].length) {
         fixed = new boolean[centers.length-1][centers[0].length-1];
         C = new double[centers.length-1][];
         for(int c=0; c<centers.length-1; c++) {
            C[c] = Arrays.copyOf(centers[c],centers[c].length-1);
            for(int i=0; i<centers[c].length-1; i++)
               fixed[c][i] = Double.isNaN(centers[c][data[0].length]) || Double.isNaN(centers[centers.length-1][i]);
            }
         }
      double[][] distsq = new double[C.length][];
      for(int c=0; c<C.length; c++)
         distsq[c] = Matrix.rowsum(Matrix.SQ(Matrix.sub(data,C[c])));
      double[][] U = Matrix.POW(distsq,-1.0/(ff-1.0));
      U = Matrix.DIVj(U, Matrix.colsum(U));
      double[][] mf = Matrix.POW(U,ff);
      double[] gof = new double[NI];
      gof[0] = Matrix.sumsum(Matrix.mul(distsq,mf));
      double[] ind = new double[NI];
      ind[0] = 0;
      for(int i=1; i<NI; i++) {
         IJ.showProgress(i/(float)NI);
         //processWidget.setLabel(""+i);
         C = Matrix.DIVi(Matrix.multiply(mf,data),Matrix.rowsum(mf));
         for(int c=0; c<C.length; c++) {
            if (fixed != null)
               for(int ii=0; ii<C[c].length; ii++)
                  if (fixed[c][ii]) C[c][ii] = centers[c][ii];
            distsq[c] = Matrix.rowsum(Matrix.SQ(Matrix.sub(data,C[c])));
            }
         gof[i] = Matrix.sumsum(Matrix.mul(distsq,mf));
         ind[i] = i;
//IJ.log(""+i+": "+gof[i]);
         if (Math.abs(gof[i]-gof[i-1]) < eps) {
            IJ.showStatus("fcm: threshold reached");
            return new double[][][]{U,C,new double[][]{Arrays.copyOfRange(gof,0,i),Arrays.copyOfRange(ind,0,i)}};
            }
         IJ.showStatus(""+Math.abs(gof[i]-gof[i-1])+" < "+eps);
         U = Matrix.POW(distsq,-1.0/(ff-1.0));
         U = Matrix.DIVj(U, Matrix.colsum(U));
         mf = Matrix.POW(U,ff);
         }
      IJ.showProgress(1.0);
      IJ.showStatus("fcm: max iterations");
      return new double[][][]{U,C,new double[][]{gof,ind}};
      }

   /** rudementory matrix utility methods
    */
   public static class Matrix {

      public static double[][] sub(double[][] x, double[] v) {
         int m1 = x.length;
         int n1 = x[0].length;
      
         double result[][] = new double[m1][n1];
         for(int i=0; i<m1; i++)
            for(int j=0; j<n1; j++)
               result[i][j] = x[i][j] - v[j];
         return result;
         }

      // return c = a * b
      public static double[][] multiply(double[][] a, double[][] b) {
         int m1 = a.length;
         int n1 = a[0].length;
         int m2 = b.length;
         int n2 = b[0].length;
         if (n1 != m2) throw new RuntimeException("Illegal matrix dimensions.");
         double[][] c = new double[m1][n2];
         for(int i = 0; i < m1; i++)
             for(int j = 0; j < n2; j++)
              for(int k = 0; k < n1; k++)
                 c[i][j] += a[i][k] * (Double.isFinite(b[k][j]) ? b[k][j] : 0.0);
         return c;
         }
/*
      // matrix-vector multiplication (y = A * x)
      public static double[] multiply(double[][] a, double[] x) {
          int m = a.length;
          int n = a[0].length;
          if (x.length != n) throw new RuntimeException("Illegal matrix dimensions.");
          double[] y = new double[m];
          for (int i = 0; i < m; i++)
              for (int j = 0; j < n; j++)
                  y[i] += a[i][j] * x[j];
          return y;
         }
*/
     public static double[][] DIVi(double[][] x, double[] y) {
         for(int i=0; i<x.length; i++)
            for(int j=0; j<x[0].length; j++) 
               x[i][j] = (y[i] != 0) ? x[i][j]/y[i] : 0;
         return x;
         }

     public static double[][] DIVj(double[][] x, double[] y) {
         for(int i=0; i<x.length; i++)
            for(int j=0; j<x[0].length; j++) 
               x[i][j] = (y[j] != 0) ? x[i][j]/y[j] : 0;
         return x;
         }

      // Array multiply
      public static double[][] mul(double[][] x, double[][] y) {
         double result[][] = new double[x.length][x[0].length];
         for(int i=0; i<x.length; i++)
            for(int j=0; j<x[0].length; j++) 
               result[i][j] = x[i][j]*y[i][j];
         return result;
         }
      // Array multiply
      public static double[][] MUL(double[][] x, double[][] y) {
         for(int i=0; i<x.length; i++)
            for(int j=0; j<x[0].length; j++) 
               x[i][j] = x[i][j]*y[i][j];
         return x;
         }

      public static double[][] SQ(double[][] x) {
         int m1 = x.length;
         int n1 = x[0].length;
      
         for(int i=0; i<m1; i++)
            for(int j=0; j<n1; j++)
               x[i][j] = x[i][j] * x[i][j];
         return x;
         }

      public static double[][] POW(double[][] x, double exponent) {
         int m1 = x.length;
         int n1 = x[0].length;
    
         double[][] result = new double[m1][n1];
         for(int i=0; i<m1; i++)
            for(int j=0; j<n1; j++)
               result[i][j] = Math.pow(Math.max(x[i][j],0.001),exponent);  // how small/big???
         return result;
         }

      public static double[] rowsum(double[][] x) {
         int m1 = x.length;
         int n1 = x[0].length;
    
         double[] result = new double[m1];
         for(int i=0; i<m1; i++)
            for(int j=0; j<n1; j++)
               result[i] += Double.isFinite(x[i][j]) ? x[i][j] : 0.0;
         return result;
         }

      public static double[] colsum(double[][] x) {
         int m1 = x.length;
         int n1 = x[0].length;
    
         double[] result = new double[n1];
         for(int i=0; i<m1; i++)
            for(int j=0; j<n1; j++)
               result[j] += Double.isFinite(x[i][j]) ? x[i][j] : 0.0;
         return result;
         }

      public static double sumsum(double[][] x) {
         double result = 0;
         for(int i=0; i<x.length; i++)
            for(int j=0; j<x[0].length; j++)
               result += Double.isFinite(x[i][j]) ? x[i][j] : 0.0;
         return result;
         }
      }

}
